/*
 * Copyright 1997-2025 Optimatika
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package org.ojalgo.matrix.store;

import org.ojalgo.scalar.Scalar;

/**
 * A selection (re-ordering) of rows.
 *
 * @author apete
 */
class RowsStore<N extends Comparable<N>> extends SelectingStore<N> {

    int[] rows;

    RowsStore(final MatrixStore<N> base, final int... selection) {

        super(base, selection.length, base.getColDim());

        rows = selection;
    }

    @Override
    public double doubleValue(final int row, final int col) {
        return this.base.doubleValue(rows[row], col);
    }

    @Override
    public int firstInRow(final int row) {
        return this.base.firstInRow(rows[row]);
    }

    @Override
    public N get(final int row, final int col) {
        return this.base.get(rows[row], col);
    }

    @Override
    public int limitOfRow(final int row) {
        return this.base.limitOfRow(rows[row]);
    }

    @Override
    public void supplyTo(final TransformableRegion<N> consumer) {
        MatrixStore<N> base = this.base;
        for (int i = 0, limit = rows.length; i < limit; i++) {
            consumer.fillRow(i, base.sliceRow(rows[i]));
        }
    }

    @Override
    public Scalar<N> toScalar(final int row, final int col) {
        return this.base.toScalar(rows[row], col);
    }

}
